from typing import Any, Callable, Dict, Sequence, Type

import xappt

from xappt.models.parameter.model import Parameter


@xappt.register_plugin
class StdIO(xappt.BaseInterface):
    def __init__(self):
        super().__init__()

        self.validation_dispatch: Dict[Type, Callable] = {
            int: self.check_int,
        }

        self.prompt_dispatch: Dict[Type, Callable] = {
            int: self.prompt_int,
        }

    def invoke(self, plugin: xappt.BaseTool, **kwargs):
        for param in plugin.parameters():
            self.prompt_for_param(param)
        return plugin.execute(**kwargs)

    def check_default(self, param: Parameter, value: Any):
        if value is None and param.default is not None:
            value = param.default
        return value

    def check_int(self, param: Parameter, value: Any):
        if param.choices is not None:
            if value is None and param.default is not None:
                value = param.choices[param.default]
            if value not in param.choices:
                raise ValueError(f"Value must be one of {xappt.humanize_list(param.choices)}")
            value = param.choices.index(value)
        else:
            if value is None and param.default is not None:
                value = param.default
            value = int(value)
            minimum = param.options.get('min')
            if minimum is not None and value < minimum:
                raise ValueError(f"Value must be at least {minimum}")
            maximum = param.options.get('max')
            if maximum is not None and value > maximum:
                raise ValueError(f"Value must be less than {maximum}")
        return value

    def prompt_default(self, param: Parameter) -> str:
        if param.default is not None:
            return f"{param.name} ({param.default}): "
        return f"{param.name}: "

    def prompt_int(self, param: Parameter) -> str:
        if param.choices is not None:
            if param.default is not None:
                choice_values = []
                for i, value in enumerate(param.choices):
                    if i == param.default:
                        choice_values.append(f'{value}*')
                    else:
                        choice_values.append(value)
            else:
                choice_values = param.choices
            return f"{param.name} ({'|'.join(choice_values)}): "
        return self.prompt_default(param)

    def prompt_for_param(self, param: Parameter):
        prompt_fn = self.prompt_dispatch.get(param.data_type, self.prompt_default)
        validation_fn = self.validation_dispatch.get(param.data_type, self.check_default)
        prompt = prompt_fn(param)

        print("")
        if len(param.description):
            print(param.description)

        while True:
            value = input(prompt)
            if len(value) == 0:
                value = None
            try:
                value = validation_fn(param, value)
            except BaseException as e:
                print(f"Error: {e}")
                print("Try again")
                print("")
            else:
                break

        param.value = value



import logging
import os
import subprocess

from collections import namedtuple
from typing import List, Sequence, Union


CommandResult = namedtuple("CommandResult", ["result", "stdout", "stderr"])

logger = logging.getLogger(__name__)


class CommandRunner(object):
    """ A wrapper around `subprocess.Popen` with basic environment manipulation.
    The results of a subprocess operation will be returned as a `CommandResult`
    object, which includes the return code, and stdout/stderr. Note that
    stdout/stderr will only be populated if the command was run silently.

    >>> c = CommandRunner(env={})
    >>> c.env_var_add("TEST", "1234")
    >>> c.env['TEST']
    '1234'
    >>> if os.name == "nt":
    ...     command = "set"
    ... elif os.name == "posix":
    ...     command = "printenv"
    >>> result = c.run(command, silent=True)
    >>> result.result == 0
    True
    >>> result.stdout.count("TEST=1234")
    1

    """
    def __init__(self, **kwargs):
        self.cwd = kwargs.get('cwd', os.getcwd())
        self.env = kwargs.get('env', os.environ.copy())

    def _split_path_var(self, key: str) -> List[str]:
        values = self.env.get(key, "").split(os.pathsep)
        return [v for v in values if len(v.strip())]

    def env_path_append(self, key: str, value: str):
        values = self._split_path_var(key)
        values.append(value)
        self.env[key] = os.pathsep.join(values)

    def env_path_prepend(self, key: str, value: str):
        values = self._split_path_var(key)
        values.insert(0, value)
        self.env[key] = os.pathsep.join(values)

    def env_var_add(self, key: str, value: str):
        self.env[key] = value

    def env_var_remove(self, key: str):
        try:
            del self.env[key]
        except KeyError:
            pass

    def run(self, command: Union[bytes, str, Sequence], **kwargs) -> CommandResult:
        env = kwargs.get('env', self.env)

        subprocess_args = {
            'cwd': kwargs.get('cwd', self.cwd),
            'env': env,
            'shell': False,
            'universal_newlines': True,
            'encoding': 'utf8',
        }

        silent = kwargs.get('silent', True)
        if silent:
            subprocess_args['stdout'] = subprocess.PIPE
            subprocess_args['stderr'] = subprocess.PIPE

        logger.debug("Running command %s", str(command))
        logger.debug("Command environment %s", str(env))
        logger.debug("Command working directory %s", subprocess_args['cwd'])
        proc = subprocess.Popen(command, **subprocess_args)
        result = proc.wait()

        if silent:
            stdout = proc.stdout.read()
            stderr = proc.stderr.read()
            proc.stdout.close()
            proc.stderr.close()
            return CommandResult(result, stdout, stderr)
        else:
            return CommandResult(result, None, None)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import argparse
import sys

import xappt

from xappt.models.parameter import convert


def cli_main(*argv) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action='store_true', dest='version', help='Display the version number')
    subparsers = parser.add_subparsers(help="Sub command help", dest='command')

    for plugin_name, plugin_class in xappt.plugin_manager.registered_tools():
        assert issubclass(plugin_class, xappt.BaseTool)
        plugin_parser = subparsers.add_parser(plugin_class.name().lower(), help=plugin_class.help())
        for parameter in plugin_class.class_parameters():
            setup_args = parameter.param_setup_args
            args, kwargs = convert.to_argument_dict(setup_args)
            plugin_parser.add_argument(*args, **kwargs)

    options = parser.parse_args(args=argv)

    if options.version:
        print(f"xappt {xappt.version_str}")
        return 0

    if options.command is not None:
        try:
            plugin_class = xappt.plugin_manager.get_tool_plugin(options.command)
        except ValueError:
            print(f"Plugin {options.command} not found")
        else:
            plugin_instance = plugin_class(**options.__dict__)
            return plugin_instance.execute()

    parser.print_help()

    return 0


def entry_point() -> int:
    xappt.discover_plugins()
    return cli_main(*sys.argv[1:])


if __name__ == '__main__':
    sys.exit(entry_point())



import abc

from xappt.models.plugins.base import BasePlugin
from xappt.models.plugins.tool import BaseTool


class BaseInterface(BasePlugin, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def invoke(self, plugin: BaseTool, **kwargs):
        pass



import errno
import os
import shutil


def _find_python_nt(version):
    import winreg

    opened_key = None
    for key in (winreg.HKEY_LOCAL_MACHINE, winreg.HKEY_CURRENT_USER):
        try:
            opened_key = winreg.OpenKey(key, r"SOFTWARE\Python\PythonCore\{0}".format(version))
        except WindowsError as e:
            if e.errno != errno.ENOENT:
                raise

    if opened_key is None:
        return None

    py_path = winreg.QueryValue(opened_key, "InstallPath")
    py_exe = shutil.which("python", path=py_path)
    if py_exe is not None and os.path.isfile(py_exe):
        return py_exe

    return None


def _find_python_posix(version):
    return shutil.which("python{0}".format(version))


def find_python(major, minor):
    """ Helper to find the executable for a specific python version.

    >>> # find Python 3.6 or higher
    >>> py_bin = find_python(3, 9) or find_python(3, 8) or find_python(3, 7) or find_python(3, 6)
    >>> assert py_bin is not None

    """
    version = "{major}.{minor}".format(**locals())
    if os.name == "nt":
        return _find_python_nt(version)
    elif os.name == "posix":
        return _find_python_posix(version)
    raise NotImplementedError



import xappt


@xappt.register_plugin
class Interactive1(xappt.BaseTool):
    param1 = xappt.ParamInt(required=True)

    def execute(self, **kwargs) -> int:
        iface_class = xappt.get_default_interface()
        iface = iface_class()
        try:
            result = iface.invoke(Interactive2(), **self.param_dict())
        except KeyboardInterrupt:
            result = 1
        return result


@xappt.register_plugin(visible=False)
class Interactive2(xappt.BaseTool):
    param1 = xappt.ParamInt(description="A number between 1 and 10, inclusive", minimum=1, maximum=10)
    param2 = xappt.ParamInt(description="A number between 10 and 20, inclusive", minimum=10, maximum=20)
    param3 = xappt.ParamInt(default=7, description="Any integer")
    param4 = xappt.ParamInt(description="Run silently?", choices=("y", "n"), default=0)

    def execute(self, **kwargs) -> int:
        print(f"Got {self.param1.value}, {self.param2.value}, {self.param3.value}, "
              f"{self.param4.choices[self.param4.value]}")
        return 0



from typing import Iterable


def humanize_list(items: Iterable, conjunction: str = "or") -> str:
    conjunction = f" {conjunction.strip()} "
    items = list(map(lambda x: str(x).strip(), items))
    if len(items) == 0:
        raise ValueError("Expected a collection with at least one item")
    if len(items) == 1:
        return items[0]
    if len(items) == 2:
        return conjunction.join(items)
    return f"{', '.join(items[:-1])}, {conjunction}{items[-1]}"


MAGNITUDE = {
    1000: ('B', 'kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'),
    1024: ('B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB'),
}


def humanize_bytes(value, *, decimal_places: int = 0, binary: bool = True) -> str:
    f = f"{{0:.{decimal_places}f}} {{1}}"
    b = 1024 if binary else 1000
    m = MAGNITUDE[b]
    for i in range(len(m)):
        c = b ** i
        if c >= value:
            p = max(0, i - 1)
            return f.format(float(value) / float(b ** p), m[p])
    p = len(m) - 1
    return f.format(value / (b ** p), m[p])



import xappt.__version__

from xappt.managers import plugin_manager
from xappt.managers.plugin_manager import *

from xappt.models import BaseTool, BaseInterface
from xappt.models.parameter.parameters import *

from xappt.utilities import *


version = tuple(map(int, xappt.__version__.__version__.split('.'))) + (xappt.__version__.__build__, )
version_str = f"{xappt.__version__.__version__}-{xappt.__version__.__build__}"



from typing import Optional, Sequence

from .model import ParameterDescriptor


__all__ = [
    'ParamString',
    'ParamBool',
    'ParamInt',
    'ParamFloat',
]


class ParamString(ParameterDescriptor):
    def __init__(self, **kwargs):
        super().__init__(data_type=str, **kwargs)


class ParamBool(ParameterDescriptor):
    def __init__(self, **kwargs):
        super().__init__(data_type=bool, **kwargs)


class ParamInt(ParameterDescriptor):
    def __init__(self, *, minimum: Optional[int] = None, maximum: Optional[int] = None,
                 choices: Optional[Sequence] = None, **kwargs):
        super().__init__(data_type=int, minimum=minimum, maximum=maximum, choices=choices, **kwargs)


class ParamFloat(ParameterDescriptor):
    def __init__(self, **kwargs):
        super().__init__(data_type=float, **kwargs)



import copy

from typing import Type


class Parameter:
    def __init__(self, name: str, *, data_type: Type, **kwargs):
        self.name = name
        self.data_type = data_type
        self.description = kwargs['description']
        self.default = kwargs['default']
        self.value = kwargs['value']
        self.required = kwargs['required']
        self.minimum = kwargs.get('minimum')
        self.maximum = kwargs.get('maximum')
        self.choices = kwargs.get('choices')
        self.options = kwargs.get('options', {})


class ParameterDescriptor:
    __counter = 0

    default_values = {
        'description': "",
        'short_name': None,
    }

    def __init__(self, data_type: Type, **kwargs):
        cls = self.__class__
        self.storage_name = '_{}#{}'.format(cls.__name__, cls.__counter)
        kwargs['data_type'] = data_type
        if 'default' not in kwargs:
            kwargs['required'] = True
        else:
            kwargs['required'] = False
        kwargs['default'] = kwargs.get('default')
        kwargs['value'] = kwargs.get('value', kwargs['default'])
        for k, v in self.default_values.items():
            if k not in kwargs:
                kwargs[k] = v
        self.param_setup_args = copy.deepcopy(kwargs)
        cls.__counter += 1

    def get_parameter(self, instance):
        param = getattr(instance, self.storage_name, None)
        if param is None:
            # Create new parameter instance of descriptor properties
            param = Parameter(**self.param_setup_args)
            setattr(instance, self.storage_name, param)
            return param
        return param

    def __get__(self, instance, owner):
        if instance is None:
            # Call was not through an instance
            return self
        else:
            return self.get_parameter(instance)

    def __set__(self, instance, value):
        raise AttributeError("Parameter is read only. Use Parameter's value attribute")
